#!/usr/bin/env python3

import click

import mucus.command
import mucus.config
import mucus.deezer.client
import mucus.deezer.exception
import mucus.exception
import mucus.history
import mucus.play.player


@click.command(help='music cli thing', epilog='\b\n' + click.style('''
 ███▄ ▄███▓ █    ██  ▄████▄   █    ██   ██████
▓██▒▀█▀ ██▒ ██  ▓██▒▒██▀ ▀█   ██  ▓██▒▒██    ▒
▓██    ▓██░▓██  ▒██░▒▓█    ▄ ▓██  ▒██░░ ▓██▄
▒██    ▒██ ▓▓█  ░██░▒▓▓▄ ▄██▒▓▓█  ░██░  ▒   ██▒
▒██▒   ░██▒▒▒█████▓ ▒ ▓███▀ ░▒▒█████▓ ▒██████▒▒
░ ▒░   ░  ░░▒▓▒ ▒ ▒ ░ ░▒ ▒  ░░▒▓▒ ▒ ▒ ▒ ▒▓▒ ▒ ░
░  ░      ░░░▒░ ░ ░   ░  ▒   ░░▒░ ░ ░ ░ ░▒  ░ ░
░      ░    ░░░ ░ ░ ░         ░░░ ░ ░ ░  ░  ░
       ░      ░     ░ ░         ░           ░
                    ░
'''.strip(), fg='green'))
@click.option('--alias', '-a', 'aliases', type=(str, str),
              metavar='<name> <command>', multiple=True)
@click.option('--default', '-d', default='search', show_default=True,
              metavar='<command>', help='default command')
def command(aliases, default):

    aliases = dict(aliases)
    config = mucus.config.Config()

    with mucus.history.History(__name__) as history:
        try:
            client = mucus.deezer.client.Client()
        except mucus.deezer.exception.AuthException as e:
            raise click.ClickException(e)
        player = mucus.play.player.Player(client)
        player.start()

        def inputs():
            while True:
                try:
                    yield input('> ')
                except EOFError:
                    break

        for line in inputs():
            if line.strip() == '':
                continue

            try:
                loader = mucus.command.Loader(line=line, aliases=aliases)
            except mucus.command.NoSuchCommand:
                try:
                    loader = mucus.command.Loader(name=default)
                except mucus.command.NoSuchCommand as e:
                    raise click.ClickException(e)

            runner = mucus.command.Runner(
                loader,
                context={
                    'client': client,
                    'command': {
                        'line': line,
                        'name': loader.name
                    },
                    'config': config,
                    'history': history,
                    'player': player
                }
            )

            try:
                runner()
            except mucus.exception.Exit:
                break
            except Exception as e:
                raise click.ClickException(e)


def main():
    return command(default_map=mucus.config.Config().flatten())



import json

import redis


class History:
    def __init__(self):
        self.redis = redis.Redis(decode_responses=True)
        self.key = ':'.join((__name__, self.__class__.__name__))

    def __repr__(self):
        return f'<{self.__class__.__name__}>'

    def append(self, song):
        self.redis.lpush(self.key, json.dumps(song, separators=(',', ':')))



import re

from mucus.exception import NoSuchCommand
from mucus.lazy import import_module as lazy_import


class Command:
    pass


class Loader:
    def __init__(self, line=None, name=None, aliases=None):
        if aliases is None:
            aliases = {}
        if line is not None:
            if line in aliases:
                name = aliases[line]
            else:
                match = re.match(r'^([A-Za-z0-9_-]+)(.*)', line)
                if match is None:
                    raise NoSuchCommand
                name = match.group(1).replace('-', '_')
        if name is None:
            raise NoSuchCommand
        try:
            name = aliases[name]
        except KeyError:
            pass
        try:
            self.mod = lazy_import('.' + name, 'mucus.commands')
        except ModuleNotFoundError as e:
            raise NoSuchCommand(e)
        self.cmd = self.mod.Command()
        self.name = name


class Runner:
    def __init__(self, loader, context):
        self.loader = loader
        self.context = context

    def __call__(self):
        return self.loader.cmd(**self.context)



import readline

import redis


class History:
    def __init__(self, namespace, n=1000, sep=':'):
        self.namespace = namespace
        self.n = n
        self.sep = sep

    def __enter__(self):
        try:
            with redis.Redis(decode_responses=True) as db:
                self.backup = [readline.get_history_item(i+1)
                               for i in range(readline.get_current_history_length())] # noqa
                readline.clear_history()
                for item in db.lrange(self.key, -self.n, -1):
                    readline.add_history(item)
                db.expire(self.key, 2592000)
        except redis.exceptions.RedisError:
            pass

    def __exit__(self, *args):
        try:
            with redis.Redis(decode_responses=True) as db:
                for i in range(readline.get_current_history_length()):
                    db.rpush(self.key, readline.get_history_item(i+1))
                db.expire(self.key, 2592000)
                readline.clear_history()
                for item in self.backup:
                    readline.add_history(item)
        except redis.exceptions.RedisError:
            pass

    @property
    def key(self):
        return self.sep.join((__name__, self.namespace))



import os

import click

import mucus.command
import mucus.commands


class Command(mucus.command.Command):
    def __call__(self, **kwargs):
        for f in os.listdir(mucus.commands.__path__[0]):
            if f.startswith('_'):
                continue
            if f.endswith('.py'):
                click.echo(f[:-3])



class Page:
    method = None
    params = None

    def __init__(self, client, **kwargs):
        self._data = client.post(self.method, json={**self.params, **kwargs})

    def __repr__(self):
        return f'<{self.__class__.__name__}>'

    def __getattr__(self, name):
        try:
            return self._data[name.upper()]
        except KeyError:
            raise AttributeError(name)


class Artist(Page):
    method = 'deezer.PageArtist'
    params = {'art_id': None, 'lang': 'en', 'tab': 0}


class Album(Page):
    method = 'deezer.PageAlbum'
    params = {'alb_id': None, 'lang': 'en', 'tab': 0, 'header': True}


class Search(Page):
    method = 'deezer.PageSearch'
    params = {
        'query': None,
        'start': 0,
        'nb': 10,
        'suggest': True,
        'artist_suggest': True,
        'top_tracks': True
    }

    def __iter__(self):
        def generator():
            for track in self.data['TRACKS']:
                yield track
        return generator()



class Song:
    def __init__(self, data):
        self.data = data

    @property
    def artist(self):
        return self.data['ART_NAME']

    @property
    def title(self):
        return self.data['SNG_TITLE']

    @property
    def track(self):
        return self.data['TRACK_NUMBER']

    @property
    def album(self):
        return self.data['ALB_TITLE']

    def __repr__(self):
        return f'<Song title={self.title!r} artist={self.artist!r}>'

    def __str__(self):
        return f'{self.artist} - {self.title}'



import rich

import mucus.command


class Command(mucus.command.Command):
    def __call__(self, player, **kwargs):
        def default(obj):
            return str(obj)
        rich.print_json(data=player.state, default=default)



import collections
import queue
import subprocess
import threading

from mucus.exception import NoMedia, NoSource
from mucus.song import Song

events = ('play', 'stop')


class Events(collections.namedtuple('Events', events)):
    def __new__(cls):
        return super().__new__(cls, *(threading.Event() for _ in events))


class Player:
    def __init__(self, client):
        self.client = client
        self.queue = queue.Queue()
        self.events = Events()
        self.thread = None
        self.song = None
        self.history = []

    def loop(self):
        self.events.stop.clear()
        while not self.events.stop.is_set():
            try:
                song = self.queue.get(timeout=1)
            except queue.Empty:
                continue
            if song is None:
                break
            self.play(song)
            self.queue.task_done()

    def play(self, song):
        self.history.append(song)

        data = self.client.stream(song)
        media = next(data)
        if media is None:
            raise NoMedia
        source = next(data)
        if source is None:
            raise NoSource

        self.events.play.set()

        with subprocess.Popen(['sox', '-', '-d', '-q', '-V0'], stdin=subprocess.PIPE) as p: # noqa
            for chunk in data:
                if self.events.stop.is_set():
                    break
                self.events.play.wait()
                p.stdin.write(chunk)

        self.events.play.clear()

    def start(self):
        self.thread = threading.Thread(target=self.loop, daemon=True)
        self.thread.start()

    @property
    def state(self):
        try:
            song = self.history[-1]
        except IndexError:
            song = None
        return {
            'playing': self.events.play.is_set(),
            'song': Song(song)
        }



import importlib.util


def import_module(name, package=None):
    spec = importlib.util.find_spec(name, package)
    if spec is None:
        raise ModuleNotFoundError(name)
    loader = importlib.util.LazyLoader(spec.loader)
    spec.loader = loader
    module = importlib.util.module_from_spec(spec)
    loader.exec_module(module)
    return module



import os

import confuse


class Config(confuse.LazyConfig):
    def __init__(self, name=None):
        super().__init__('mucus', __name__)
        if name is not None:
            filename = os.path.join(self.config_dir(), name)
            if not filename.endswith('.yaml'):
                filename += '.yaml'
            self.set_file(filename)



import copy
import json

import click
import yaspin

import myass.assistant

import mucus.command
import mucus.deezer.page
import mucus.exception
import mucus.history
import mucus.song


class Assistant(myass.assistant.Assistant):
    def __init__(self):
        super().__init__('mucus')
        self.context = {}

    def __call__(self, content):
        params = copy.deepcopy(self.config.flatten())
        self.messages.append({'role': 'user', 'content': content})
        params['messages'].extend(self.messages)
        with yaspin.yaspin(timer=True):
            r = self.api.post('chat/completions', json=params)
        m = r['choices'][0]['message']
        self.messages.append(m)
        params['messages'].append(m)
        while m.get('function_call') is not None:
            fn = m['function_call']['name']
            try:
                f = getattr(self, fn)
            except AttributeError:
                r = None
            else:
                r = f(**json.loads(m['function_call']['arguments']))
            m = {'role': 'function', 'name': fn, 'content': str(r)}
            params['messages'].append(m)
            with yaspin.yaspin(timer=True):
                r = self.api.post('chat/completions', json=params)
            params['messages'].pop()
            m = r['choices'][0]['message']
            self.messages.append(m)
        yield m['content']

    def play_song(self, artist, title):
        client = self.context['client']
        query = f'{artist} {title}'
        search = mucus.deezer.page.Search(client=client, query=query, nb=1)
        song = mucus.song.Song(search.track['data'][0])
        self.context['player'].queue.put(song)
        return str(song)

    def play_songs(self, songs):
        client = self.context['client']
        for song in songs:
            query = ' '.join((song['artist'], song['title']))
            search = mucus.deezer.page.Search(client=client, query=query, nb=1)
            song = mucus.song.Song(search.track['data'][0])
            self.context['player'].queue.put(song)

    def exit(self):
        raise mucus.exception.Exit


class Command(mucus.command.Command):
    def __init__(self):
        self.assistant = Assistant()

    def __call__(self, **kwargs):
        assist = self.assistant
        assist.context.update(**kwargs)
        with mucus.history.History(__name__):
            while True:
                try:
                    inputs = click.prompt(':', prompt_suffix=' ')
                except click.exceptions.Abort:
                    break
                try:
                    for content in assist(inputs):
                        click.echo(content)
                except mucus.exception.Exit:
                    break



import os

import mucus.command


class Command(mucus.command.Command):
    def __call__(self, client, config, player, **kwargs):
        song = None
        for k in ('song', 'last_song'):
            try:
                song = player.state[k]
            except KeyError:
                continue
        if song is None:
            return
        data = client.stream(song)
        media = next(data)
        if media is None:
            raise Exception
        source = next(data)
        if source is None:
            raise Exception
        root = config['download']['directory'].get()
        for path in (os.path.join(root, song.artist),
                     os.path.join(root, song.artist, song.album)):
            os.mkdir(path)
        dn = os.path.join(config['download']['directory'].get(),
                          song.artist,
                          song.album)
        fn = f'{song.track} {song.title}.flac'
        with open(os.path.join(dn, fn), 'wb') as f:
            for chunk in data:
                f.write(chunk)



import click

import mucus.command
import mucus.deezer.page
import mucus.history


class Command(mucus.command.Command):
    def __call__(self, client, command, player, **kwargs):
        query = command['line']
        search = mucus.deezer.page.Search(client=client, query=query, nb=20)
        choices = search.track['data'].copy()
        for i, track in enumerate(choices):
            click.echo(' '.join([
                click.style(f'{i:03}', fg='red'),
                click.style(track['ART_NAME'], fg='green'),
                click.style(track['SNG_TITLE'], fg='blue')
            ]))
        with mucus.history.History(__name__):
            try:
                i = input('# ')
            except EOFError:
                return
        if ':' in i:
            i = slice(*map(lambda x: x.isdigit() and int(x) or None, i.split(':')))
        else:
            try:
                i = int(i)
            except ValueError:
                return
            i = slice(i, i+1)
        for choice in choices[i]:
            player.queue.put(choice)



import redis

from mucus.deezer.exception import AuthException


def discover(key):
    try:
        with redis.Redis(decode_responses=True) as db:
            value = db.get(f'mucus:deezer:{key}')
            if value is not None:
                return value
    except redis.exceptions.RedisError:
        pass
    raise AuthException(key)


class Auth:
    def __init__(self, sid=None, arl=None):
        if sid is None:
            sid = discover('sid')
        if arl is None:
            arl = discover('arl')
        self.sid = sid
        self.arl = arl



import httpx

from mucus.deezer.decrypter import Decrypter


class Stream:
    def __init__(self, song, license_token):
        try:
            song = song.data
        except AttributeError:
            pass
        self.song = song
        self.license_token = license_token

    def stream(self, format='FLAC'):
        media = httpx.post('https://media.deezer.com/v1/get_url', json={
            'license_token': self.license_token,
            'media': [{'type': 'FULL',
                       'formats': [{'cipher': 'BF_CBC_STRIPE',
                                    'format': format}]}],
            'track_tokens': [self.song['TRACK_TOKEN']]
        }).json()['data'][0]['media'][0]
        yield media
        source = media['sources'][0]
        yield source
        decrypter = Decrypter(self.song['SNG_ID'])
        url = source['url']
        pos = 0
        while True:
            try:
                with httpx.stream('GET', url, headers={'range': f'bytes={pos}-'}, timeout=1.0) as r: # noqa
                    for chunk in r.iter_bytes(chunk_size=2048):
                        if (pos // 2048) % 3 == 0 and len(chunk) == 2048:
                            chunk = decrypter.decrypt(chunk)
                        yield chunk
                        pos += len(chunk)
                    return
            except httpx.HTTPError:
                continue



import queue

import click

import mucus.command


class Command(mucus.command.Command):
    def __call__(self, command, player, **kwargs):
        items = []
        while True:
            try:
                items.append(player.queue.get(block=False))
            except queue.Empty:
                break
        if '!' in command['line']:
            return
        for i, item in enumerate(items):
            click.echo(' '.join([
                click.style(f'{i:03}', fg='red'),
                click.style(item['ART_NAME'], fg='green'),
                click.style(item['SNG_TITLE'], fg='blue')
            ]))
            player.queue.put(item)



import mucus.command


class Command(mucus.command.Command):
    def __call__(self, player, **kwargs):
        if player.events.play.is_set():
            player.events.play.clear()
        else:
            player.events.play.set()



import hashlib

from Crypto.Cipher import Blowfish as bf


class Decrypter:
    secret = b'g4el58wc0zvf9na1'
    iv = bytes(range(8))

    def __init__(self, song_id):
        h = hashlib.md5(song_id.encode()).hexdigest().encode()
        self.key = bytes([h[i] ^ h[i+16] ^ self.secret[i] for i in range(16)])

    def decrypt(self, chunk):
        return bf.new(self.key, bf.MODE_CBC, self.iv).decrypt(chunk)



import mucus.command


class Command(mucus.command.Command):
    def __call__(self, player, **kwargs):
        if player.thread is not None:
            player.events.stop.set()
            player.thread.join()
        player.start()



import random

import httpx

from mucus.deezer.stream import Stream
from mucus.deezer.auth import Auth
from mucus.deezer.exception import ApiException, AuthException


class Client(httpx.Client):
    def __init__(self, sid=None, arl=None, user_data=None):
        super().__init__(base_url='https://www.deezer.com/ajax/')

        auth = Auth(sid, arl)
        self.cookies['sid'] = auth.sid
        self.cookies['arl'] = auth.arl
        self.user = {}
        self.user.update(self.post('deezer.getUserData'))
        if self.user['USER']['USER_ID'] <= 0:
            raise AuthException

    @property
    def license_token(self):
        try:
            return self.user['USER']['OPTIONS']['license_token']
        except KeyError:
            return None

    def request(self, method, url, **kwargs):
        kwargs['params'] = {
            'method': url,
            'input': 3,
            'api_version': '1.0',
            'api_token': self.user.get('checkForm', ''),
            'cid': random.randrange(0, 1000000000)
        }
        r = super().request(method, 'gw-light.php', **kwargs)
        r.raise_for_status()
        r = r.json()
        if r['error']:
            raise ApiException(r['error'])
        return r['results']

    def stream(self, song, format='FLAC'):
        return Stream(song, self.license_token).stream(format)
